package org.jboss.repro.stateful.timeout;

import java.util.Arrays;
import java.util.Properties;

import javax.ejb.EJBObject;

import org.jboss.arquillian.container.test.api.ContainerController;
import org.jboss.arquillian.container.test.api.Deployer;
import org.jboss.arquillian.container.test.api.Deployment;
import org.jboss.arquillian.container.test.api.RunAsClient;
import org.jboss.arquillian.container.test.api.TargetsContainer;
import org.jboss.arquillian.junit.Arquillian;
import org.jboss.arquillian.test.api.ArquillianResource;
import org.jboss.ejb.client.Affinity;
import org.jboss.ejb.client.ClusterAffinity;
import org.jboss.ejb.client.ContextSelector;
import org.jboss.ejb.client.EJBClient;
import org.jboss.ejb.client.EJBClientConfiguration;
import org.jboss.ejb.client.EJBClientContext;
import org.jboss.ejb.client.NodeAffinity;
import org.jboss.ejb.client.PropertiesBasedEJBClientConfiguration;
import org.jboss.ejb.client.StatelessEJBLocator;
import org.jboss.ejb.client.remoting.ConfigBasedEJBClientContextSelector;
import org.jboss.shrinkwrap.api.ShrinkWrap;
import org.jboss.shrinkwrap.api.spec.EnterpriseArchive;
import org.jboss.shrinkwrap.api.spec.JavaArchive;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A small reproduction scenario showing that SFSBs returned from a SLSB-facade to the client, uses the strong URI-affinity,
 * instead of relying on a {@link ClusterAffinity} as the strong affinity and the {@link NodeAffinity} as it's weak affinity.
 */
@RunWith(Arquillian.class)
public class StatefulTimeoutManagerTest {
    private static final Logger log = LoggerFactory.getLogger(StatefulTimeoutManagerTest.class);
    private static final String EAR_NAME = "ear.ear";
    private static final String JAR_NAME = "ejb.jar";

    private EJBClientContext initialContext;

    @ArquillianResource
    private ContainerController controller;
    @ArquillianResource
    private Deployer deployer;

    @Deployment(name = "deployment-1", managed = false)
    @TargetsContainer("node-1")
    private static EnterpriseArchive deployment1() {
        return createEarDeployment();
    }

    @Deployment(name = "deployment-2", managed = false)
    @TargetsContainer("node-2")
    private static EnterpriseArchive deployment2() {
        return createEarDeployment();
    }

    private static EnterpriseArchive createEarDeployment() {
        return ShrinkWrap.create(EnterpriseArchive.class, EAR_NAME).addAsModule(createEjbDeployment());
    }

    private static JavaArchive createEjbDeployment() {
        return ShrinkWrap.create(JavaArchive.class, JAR_NAME).addPackage(StatefulTimeoutEjb.class.getPackage());
    }

    /**
     * Select the preferred node, if available from the array of String-arrays.
     *
     * @param preferred
     * @param valuesSources
     * @return
     */
    private static final String prefer(final String preferred, final String[]... valuesSources) {
        for (final String[] values : valuesSources) {
            if (values == null || values.length == 0) {
                continue;
            }

            for (final String value : values) {
                if (preferred.equalsIgnoreCase(value)) {
                    log.info("Preferred: {} from: {}", preferred, Arrays.asList(values));
                    return preferred;
                }
            }

            return values[0];
        }

        throw new IllegalStateException("Unable to select a node");
    }

    @Before
    public void init() {
        // Start and deploy
        controller.start("node-1");
        controller.start("node-2");
        deployer.deploy("deployment-1");
        deployer.deploy("deployment-2");

        // store the initial EJBClientContext
        initialContext = EJBClientContext.getCurrent();

        Properties p = new Properties();
        p.put("remote.connections", "node1,node2");
        p.put("remote.connection.node1.port", "9080");
        p.put("remote.connection.node1.host", "localhost");
        p.put("remote.connection.node2.port", "10080");
        p.put("remote.connection.node2.host", "localhost");
        p.put("remote.connectionprovider.create.options.org.xnio.Options.SSL_ENABLED", "false"); // the server defaults to
                                                                                                 // SSL_ENABLED=false

        EJBClientConfiguration cc = new PropertiesBasedEJBClientConfiguration(p);
        ContextSelector<EJBClientContext> selector = new ConfigBasedEJBClientContextSelector(cc);
        EJBClientContext.setSelector(selector);

        // // Configure EJBClientContext to connect to the two nodes, preferring the first node
        // final EJBClientContext context = new EJBClientContext.Builder().
        // setDeploymentNodeSelector((eligibleNodes, appName, moduleName, distinctName) -> prefer("node-1", eligibleNodes)).
        // addClientCluster(new EJBClientCluster.Builder().
        // setName("ejb").
        // setClusterNodeSelector((clusterName, connectedNodes, totalAvailableNodes) -> prefer("node-1", totalAvailableNodes)).
        // build()).
        // addClientConnection(new EJBClientConnection.Builder().
        // setDestination(URI.create("remote+http://localhost:9080")).
        // setForDiscovery(true).build()).
        // addClientConnection(new EJBClientConnection.Builder().
        // setDestination(URI.create("remote+http://localhost:10080")).
        // setForDiscovery(true).build()).
        // addTransportProvider(new RemoteTransportProvider()).
        // build();
        // EJBClientContext.getContextManager().setGlobalDefault(context);
    }

    @After
    public void teardown() {
        // Clean up and shut down
        EJBClientContext.setSelector(initialContext.getSelector());
        deployer.undeploy("deployment-1");
        deployer.undeploy("deployment-2");
        controller.stop("node-1");
        controller.stop("node-2");
    }

    private <T extends EJBObject> Affinity getAffinity(T manager) {
        return EJBClient.getLocatorFor(manager).getAffinity();
    }
    
    @Test
    @TargetsContainer("node-1")
    @RunAsClient
    public void create() throws Exception {
        try {
            // Force the locator to use a strong cluster affinity
            final ClusterAffinity clusterAffinity = new ClusterAffinity("ejb");
            final StatelessEJBLocator<StatefulTimeoutManagerRemote> locator = new StatelessEJBLocator<>(
                    StatefulTimeoutManagerRemote.class, "ear", "ejb", "StatefulTimeoutManagerEjb", "", clusterAffinity);
            
            StatelessEJBLocator<StatefulTimeoutManagerRemote> locatorAffinityNone = new StatelessEJBLocator<>(StatefulTimeoutManagerRemote.class, "ear", "ejb", "StatefulTimeoutManagerEjb", ""); 
            
            final StatefulTimeoutManagerRemote manager = EJBClient.createProxy(locatorAffinityNone);

            log.info("Manager: {}", manager);

            
            
            // Expecting strong cluster affinity with weak affinity none
//        log.info("Manager strong affinity: {}, weak affinity: {}", getStrongAffinity(manager), getWeakAffinity(manager));
            log.info("Manager affinity: {}", getAffinity(manager));

            final StatefulTimeoutRemote statefulTimeout = manager.createStatefulTimeout();

            // // Should be strong clustered affinity with a weak node affinity, but returns URI-affinity for the node for both
            // log.info("Stateful strong affinity: {}, weak affinity: {}", getStrongAffinity(statefulTimeout),
            // getWeakAffinity(statefulTimeout));
            //
            // Ensure second node should be targeted
            // deployer.undeploy("deployment-1");

            /**
             * A strong cluster affinity must be set on the SFSB to ensure a proper failover. Uncomment the next line to emulate
             * proper strong affinity returned by the server.
             */
            // setStrongAffinity(statefulTimeout, clusterAffinity);

            log.info("Pinged node: {}", statefulTimeout.ping());
            
//        log.info("Stateful strong affinity: {}, weak affinity: {}", getStrongAffinity(statefulTimeout), getWeakAffinity(statefulTimeout));
            log.info("Stateful affinity: {}", getAffinity(statefulTimeout));
        } catch (Exception e) {
            e.printStackTrace();
            throw e;
        }
    }
}
